import os, sys, locale, io
from functools import reduce
from .table import Table
from .context import Context
from .flex import Flex
from .text import Text
from .line import Just
from .pretty import pretty
from .sgr import SGR
from .hr import HR
from .rich import Rich
from .card import Card
from .progress import Progress
from .widget import INFINITY


def stack_enum(txt, stack):
    a, b = 0, 0
    _pop, stk = [], []
    while a < len(txt):
        if a != b:
            yield txt[a:b], stk
            a = b
        # Apply pop styles to stk
        while _pop and _pop[-1].e == b:
            j, s = 0, _pop.pop().style
            while not s: j, s = j+1, _pop[-j-1].style
            stk.append(_pop[-2*(j-1)-1].style) if j else stk.pop()
        # Apply pre styles to stk 
        while stack and stack[0].s == b:
            _pop.append(stack.pop(0))
            stk.append(_pop[-1].style) if _pop[-1].style else stk.pop()
        # Advance to the next style boundry
        b = min(stack[0].s if stack else INFINITY, _pop[-1].e if _pop else INFINITY)



def _stack_enum(txt, stack):
    # De-duplicate styles
    ps, pstk = '', None
    for s, stk in _stack_enum(txt,stack):
        if pstk == None or pstk == stk:
            ps += s
        else:
            yield ps, pstk
            ps = s
        pstk = stk[:]
    yield ps, pstk



class Printer(Context):
    def __init__(self, **kwargs):
        self._widgets = []
        self.blank = 0
        super().__init__(**kwargs)


    def append(self, widget):
        self._widgets.append(widget)


    def padded(self, pad, el):
        if isinstance(pad,int):
            pad = (-pad, 0) if pad < 0 else (pad,pad)
        padl, padr = pad
        while (padl:=padl-1) >= 0:
            self.append(Text(' ',parent=self))
        self.append(el)
        while (padr:=padr-1) >= 0:
            self.append(Text(' ',parent=self))
        return self


    def __call__(self, *args, pad=0, **kwargs):
        if pad: return self.padded(pad, Text(*args, parent=self, **kwargs))
        self.append(Text(*args, parent=self, **kwargs))
        return self


    def flex(self, *args, pad=0, **kwargs):
        if pad: return self.padded(pad, Flex(*args, parent=self, **kwargs))
        self.append(Flex(*args, parent=self, **kwargs))
        return self


    def card(self, *args, pad=0, **kwargs):
        if pad: return self.padded(pad, Card(*args, parent=self, **kwargs))
        self.append(Card(*args, parent=self, **kwargs))
        return self


    def hr(self, *args, pad=0, **kwargs):
        if pad: return self.padded(pad, HR(*args, parent=self, **kwargs))
        self.append(HR(*args, parent=self, **kwargs))
        return self


    def pretty(self, *args, pad=0, **kwargs):
        for arg in args:
            self(pretty(arg,**kwargs), pad=pad, **kwargs)
        return self
        

    def calc_width(self):
        return max(0,0, *map(w.width, self._widgets))


    def flatten(self, w=0, h=0, blank=0, **kwargs):
        self.blank = blank
        if h != 0: return flatten_fix_height(w=w, h=h, **kwargs)
        if not w and Just(self['justify'],'<').h != '<':
            w = self.width # We need to be able to right/center justify to our natural width
        for widget in self._widgets:
            new_blank = 0
            for line in widget.flatten(w=w, **kwargs):
                if (is_blank:=line.is_blank()) and self.blank:
                    self.blank -= 1
                    continue
                self.blank = 0
                new_blank = new_blank+1 if is_blank else 0
                yield line
            self.blank += new_blank


    def flatten_fix_height(self, *, w, h, **kwargs):
        raise NotImplementedError()


    def __repr__(self):
        return f"<Printer>"



class Flattener(Printer):
    ''' This is an object for aliasing the print function.
    '''
    default_styles = {
        'err' : 'r!,',
        'warn': 'y!,',
        'em'  : '!',
        'dem' : 'w.;',
        '1' : 'y,',
        '2' : 'm,',
        '3' : 'c,',
    }

    print = print # The legacy print() function

    def __init__(self, *, stream=None, color=None, width=None, isatty=None, styles=default_styles, **kwargs):
        self.styles = styles
        self.stream = stream or sys.stdout
        kwargs = {k:v for k,v in kwargs.items() if v != None}
        if isatty==None:
            try:    self.isatty = self.stream.isatty()
            except: self.isatty = False
        #if 'lang' not in kwargs:
        #    kwargs['lang'] = locale.getdefaultlocale()[0]
        #kwargs['lang'] = kwargs['lang'].lower()
        if width == None:
            try:    width = os.get_terminal_size().columns-1
            except: width = INFINITY
        kwargs['width_max'] = width
        if 'ascii' not in kwargs:
            kwargs['ascii'] = (locale.getdefaultlocale()[1].lower() != 'utf-8')
        self.color = self.isatty if color == None else color
        super().__init__(**kwargs)
        self.blank = 0


    def format_out(self, txt, styles):
        stripped = txt.rstrip()
        if not self.color: return stripped
        s = ''
        sgr_prev = SGR()
        for t,stk in stack_enum(txt, styles):
            sgrs = [SGR(self.styles.get(y,y)) for y in stk] or [SGR()]
            sgr_next = reduce(lambda a,b: a+b, sgrs)
            code = sgr_next.diff(sgr_prev)
            s += code
            sgr_prev = sgr_next
            s += stripped[:len(t)]
            stripped = stripped[len(t):]
        if sgr_prev: s += '\033[0m'
        return s


    def append(self, widget):
        super().append(widget)
        for line in self.flatten(blank=self.blank):
            self.stream.write(self.format_out(*line.styled()))
            self.stream.write('\n')
        self._widgets = []


    #def each_line(self, *args, **kwargs):
    #    t = Text(*args, parent=self, **kwargs)
    #    yield from t.flatten(**kwargs)


    def progress(self, *args, **kwargs):
        return Progress(*args, parent=self, **kwargs)
        

    def to_str(self, *args, **kwargs):
        saved = self.stream
        self.stream = io.StringIO()
        self(*args, **kwargs)
        val = self.stream.getvalue()
        self.stream = saved
        return val



import functools, traceback
from contextlib import suppress
from .table import Table
from .text import Text


class Continue(Exception):pass

def pretty(val, depth=-1, _depth=0, quote=False, **kwargs):
    if depth == 0: return repr(val)
    if isinstance(val, bytes): return repr(val)
    if isinstance(val, str): return Text(val) if not quote and (q:=val.strip()) and q==val else repr(val)
    kwargs['quote'] = quote
    with suppress(Continue): return _pretty_pretty(val, _depth, depth, **kwargs)
    with suppress(Continue): return _pretty_tuple(val, _depth, depth, **kwargs)
    with suppress(Continue): return _pretty_list(val, _depth, depth, **kwargs)
    with suppress(Continue): return _pretty_dict(val, _depth, depth, **kwargs)
    with suppress(Continue): return _pretty_iter(val, _depth, depth, **kwargs)
    return repr(val)



def _pretty_pretty(val, _depth, depth, **kwargs):
    try:
        return functools.partial(val.__pretty__, _depth=_depth, depth=depth, **kwargs)()
    except Exception as e:
        if len(traceback.extract_tb(e.__traceback__)) > 1: raise e
    raise Continue()



def _pretty_tuple(val, _depth, depth, **kwargs):
    if not isinstance(val, tuple) or not val: raise Continue()
    return _pretty_enum(val, '()', _depth, depth, **kwargs)



def _pretty_list(val, _depth, depth, **kwargs):
    if not isinstance(val, list) or not val: raise Continue()
    return _pretty_enum(val, '[]', _depth, depth, **kwargs)



def _pretty_dict(val, _depth, depth, tmpl='', quote=False, **kwargs):
    try:
        v = next(iter(val.items()))
        assert(isinstance(v, tuple) and len(v) == 2)
    except:
        raise Continue()
    t = Table(0.0, 0, tmpl=tmpl, **kwargs)
    t.cell('C0', style=str(_depth%3 + 1), just='>')
    for k,v in val.items():
        if not quote and isinstance(k, str):
            k = k if (q:=k.strip()) and q==k else repr(k)
        else:
            k = repr(k)
        t(k,' \t', pretty(v, _depth=_depth+1, depth=depth-1, quote=quote, **kwargs),'\t')
    return t



def _pretty_iter(val, _depth, depth, **kwargs):
    try:
        assert(hasattr(val, '__iter__'))
        next(iter(val))
    except:
        raise Continue()
    return _pretty_enum(val, '<>', _depth, depth, **kwargs)



def _pretty_enum(val, paren, _depth, depth, tmpl='', **kwargs):
    t = Table(0.0, 0, tmpl=tmpl, **kwargs)
    t.cell('C0', style=str(_depth%3 + 1), just='>')
    for i,v in enumerate(val):
        t(f'\bdem {paren[0]}', i, f'\bdem {paren[1]} \t', pretty(v, _depth=_depth+1, depth=depth-1, **kwargs), '\t')
    return t



import pytest
from print_ext import pretty, PrettyException, Table, HR
from .testutil import printer, tostr, styled
from print_ext.line import SMark as SM



def test_pretty_str_bytes():
    v = tostr(pretty(['33', b'12', u'bob', r'\\']))
    assert(v == "[0] 33\n[1] b'12'\n[2] bob\n[3] \\\\\n")



def test_pretty_HasPretty():
    class HasPretty():
        def __init__(self, **kwargs):
            self.vals = kwargs
        
        def __pretty__(self, _depth=None, depth=None, **kwargs):
            return f"{_depth}_{depth}"

    class SubPretty(HasPretty):
        pass

    v = tostr(pretty([HasPretty(), (SubPretty(),)]))
    assert(v == '[0] 1_-2\n[1] (0) 2_-3\n')



def test_pretty_FalsePretty():
    class FalsePretty():
        def __pretty__(self): pass
        def __repr__(self): return 'FalsePretty'
    v = tostr(pretty(FalsePretty()))
    assert(v == 'FalsePretty\n')



def test_pretty_RaisesPretty():
    class SomeException(Exception): pass
    class RaisesPretty():
        def __pretty__(self, **kwargs): raise SomeException()
    with pytest.raises(SomeException):
        pretty(RaisesPretty())



def test_pretty_tuple():
    def f(*args):
        return pretty(args)
    v = tostr(f('asdf', tuple(), (b'', None)))
    assert(v == "(0) asdf\n(1) ()\n(2) (0) b''\n    (1) None\n")



def test_pretty_list():
    class MyList(list):
        pass
    v = tostr(pretty(['x', MyList([10,""])]))
    assert(v == "[0] x\n[1] [0] 10\n    [1] ''\n")



def test_pretty_dict():
    class BadDict():
        def items(self): return [1,2,3]
        def __repr__(self): return 'BadDict'
    class GoodDict():
        def items(self): return [(2,' '), (3,'\t')]
    assert(tostr(pretty({('tuple',3):BadDict()})) == "('tuple', 3) BadDict\n")
    assert(tostr(pretty({'':'multi\n\nline', 'xy':GoodDict()})) == "'' multi\n\n   line\nxy 2 ' '\n   3 '\\t'\n")



def test_pretty_multiple():
    o,p = printer()
    p.pretty(None, 'b', 33, pad=1)
    assert(o.getvalue() == "\nNone\n\nb\n\n33\n\n")



def test_pretty_quote():
    assert(tostr(pretty({'':''})) == "'' ''\n")
    assert(tostr(pretty({'\t':'\t'})) == "'\\t' '\\t'\n")
    assert(tostr(pretty({' ':'  '})) == "' ' '  '\n")
    assert(tostr(pretty({' asdf':'  asdf  '})) == "' asdf' '  asdf  '\n")
    assert(tostr(pretty({'\t':'\t '})) == "'\\t' '\\t '\n")
    assert(tostr(pretty({'\n\n':'\n\t\n'})) == "'\\n\\n' '\\n\\t\\n'\n")
    assert(tostr(pretty({'bob':'33'}, quote=True)) == "'bob' '33'\n")
    assert(tostr(pretty({'bob':'33'})) == "bob 33\n")



def test_pretty_iter():
    class Getitem():
        def __getitem__(self): return 'hi'
        def __repr__(self): return 'Getitem'
    assert(tostr(pretty(Getitem())) == 'Getitem\n')

    class Iter():
        def __iter__(self): return iter([' f ','22'])
    assert(tostr(pretty(Iter())) == "<0> ' f '\n<1> 22\n")
 


def test_pretty_exception():
    class MyIter():
        def __iter__(self):
            self.i = 0
            return self

        def __next__(self):
            if self.i == 4: raise StopIteration()
            self.i += 1
            return ['zero','one','two','three','four'][self.i-1]

        def __repr__(self):
            return 'MyIter'


    o,p = printer()
    try:
        raise PrettyException(a=MyIter(), b={'x':'hi','Hello': 'World'}, msg='Something wicked')
    except PrettyException as e:
        p(pretty(e))
        print(o.getvalue())
        assert(str(e) == 'Something wicked')
        assert(repr(e) == "PrettyException(a=MyIter, b={'x': 'hi', 'Hello': 'World'}, msg='Something wicked')")
        assert(o.getvalue() == 'Something wicked\n\na <0> zero\n  <1> one\n  <2> two\n  <3> three\nb     x hi\n  Hello World\n')



def test_pretty_exception_no_pretty():
    class MyException(PrettyException):
        def __pretty__(self, **kwargs):
            return ''
    o,p = printer()
    try:
        raise MyException(a=[1,2,3], b={'x':'hi','Hello': 'World'})
    except MyException as e:
        p(pretty(e))
        print(o.getvalue())
        assert(str(e) == "")
        assert(repr(e) == "MyException(a=[1, 2, 3], b={'x': 'hi', 'Hello': 'World'})")
        assert(o.getvalue() == '')



def test_pretty_exception_no_style():
    class MyException(PrettyException):
        def __pretty__(self, **kwargs):
            return HR('\b1 hello')
    o,p = printer(ascii=False, color=True)
    try:
        raise MyException()
    except MyException as e:
        p(pretty(e))
        assert(str(e) == "-[ hello ]-")        
        assert(repr(e) == "MyException()")
        assert(o.getvalue() == '─┤ \x1b[33mhello\x1b[0m ├─\n')



def test_pretty_styles():
    s = styled(pretty({'a':[(['b'],)]}))
    assert(s == [('a [0] (0) [0] b', [
        SM('1',0,2),
        SM('2',2,6), SM('dem',2,3), SM('dem',4,6),
        SM('3',6,10), SM('dem',6,7), SM('dem',8,10),
        SM('1',10,14), SM('dem',10,11), SM('dem',12,14)
    ])])
    


def test_pretty_depthstop():
    assert(tostr(pretty({'a':3}, depth=0)) == "{'a': 3}\n")
    assert(tostr(pretty({'a':[1,2,3]}, depth=1)) == "a [1, 2, 3]\n")



from .printer import Flattener, Printer
from .text import Text
from .table import Table
from .pretty import pretty


class PrettyException(Exception):
    ''' This implements the pretty() method to show a pretty version of the exception.
    
    __str__ and __repr__ return normal, not-pretty, strings.
    '''
    def __init__(self, **kwargs):
        for k,v in kwargs.items(): setattr(self, k, v)


    def __str__(self):
        w = self.msg if hasattr(self, 'msg') else pretty(self)
        return Flattener(ascii=True, color=False).to_str(w).rstrip()


    def __repr__(self):
        s = f"{self.__class__.__name__}("
        args = [f'{k}={v!r}' for k,v in self.__dict__.items()]
        return s+ ', '.join(args) + ')'


    def __pretty__(self, **kwargs):
        p = Printer()
        if hasattr(self, 'msg'):
            p(self.msg)
        else:
            p(self.__class__.__name__)
        vars = {k:v for k,v in self.__dict__.items() if k != 'msg'}
        if vars: p.pretty(vars, pad=-1, **kwargs)
        return p
